#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include "basemod.h"

int subtree_init(subtree_t *st, int size, int *sudoku){
	int bs,i,free;

	bs = sqrt(size);
	if(bs*bs != size){
		printf("Error in subtree_init: non square size\n");
		return PB_ERROR;
	}

	st->size = size;
	st->sudoku =  (int*) malloc(sizeof(int)*size*size);

	free = 0;
	for(i=0;i<size*size;i++){
		st->sudoku[i] = sudoku[i];
		if(sudoku[i] == 0)
			free++;
	}
	st->free = free;
	st->pos = (int*) malloc(sizeof(int)*free);
	st->upper = (int*) malloc(sizeof(int)*free);
	st->lower = (int*) malloc(sizeof(int)*free);

	free = 0;
	for(i=0;i<size*size;i++)
		if(sudoku[i] == 0)
			st->pos[free++] = i;

	for(i=0;i<st->free;i++){
		st->lower[i] = 1;
		st->upper[i] = size+1;
	}

	st->clevel = 0;
	st->sudoku[st->pos[st->clevel]] = 1;
	
	return PB_OK;
}

int subtree_destroy(subtree_t *st){
	free(st->sudoku);
	st->sudoku = NULL;
	free(st->pos);
	st->pos = NULL;
	free(st->upper);
	st->upper = NULL;
	free(st->lower);
	st->lower = NULL;

	return PB_OK;
}

/* Helper functions*/



/*advances the tree to the next*/
static int next_level(subtree_t *st){
	if(st->clevel >= st->free - 1){
//		printf("Error in next_level: already at deepest level\n");
		return PB_ERROR;
	}
	st->clevel++;
	st->sudoku[st->pos[st->clevel]] = st->lower[st->clevel];
	return PB_OK;
}

/*increases current number at current level*/
static int advance(subtree_t *st){
	if(st->sudoku[st->pos[st->clevel]] == st->upper[st->clevel] - 1){
	//	printf("Error in advance: already at upper bound at level %d\n",st->clevel);
		return PB_ERROR;
	}
	st->sudoku[st->pos[st->clevel]]++;
	return PB_OK;
}

/*tracks back one level*/
static int backtrack(subtree_t *st){
	if(st->clevel <= 0){
//		printf("Error in backtrack: already at top level\n");
		return PB_EMPTY;
	}
	st->sudoku[st->pos[st->clevel]] = 0;
	st->clevel--;
	while(advance(st) == PB_ERROR && st->clevel >= 0){
		st->sudoku[st->pos[st->clevel]] = 0;
		st->clevel--;
	}

	if(st->clevel >= 0){
		return PB_OK;
	} else {
		return PB_EMPTY;
	}
}

/*returns the number of untouched possibilities on level level*/
static int possibilities(subtree_t *st, int level){
	if(level > st->free || level < 0){
//		printf("Error in possbilities: invalid level %d\n",level);
		return PB_ERROR;
	}
	return st->upper[level] - (st->sudoku[st->pos[level]] + 1);
}

static int isempty(subtree_t *st){
	int i;
	for(i=0;i<st->free;i++)
		if(possibilities(st,i) > 0)
			return 0;
	return 1;
}

int subtree_copy(subtree_t *stin, subtree_t *stout){
	int i;

	if(stin->size != stout->size){
		printf("Error in subtree_copy: incompatible sizes\n");
		return PB_ERROR;
	}

	if(stin->free != stout->free){
		printf("Error in subtree_copy: incompatible number of free fields: in: %d vs out: %d\n",stin->free,stout->free);
		return PB_ERROR;
	}
	for(i=0;i<stin->size*stin->size;i++)
		stout->sudoku[i]=stin->sudoku[i];
	for(i=0;i<stin->free;i++){
		stout->pos[i]=stin->pos[i];
		stout->upper[i]=stin->upper[i];
		stout->lower[i]=stin->lower[i];
	}
	stout->clevel=stin->clevel;
	return PB_OK;
}
	

/*prints the current state of the subtree to stdout*/
int subtree_print(subtree_t *st){
	int i;
	printf("Clevel: %d Cupper: %d Clower: %d Valid: %d\n",st->clevel,st->upper[st->clevel],st->lower[st->clevel],subtree_check(st));
	for(i=0;i<st->size*st->size;i++){
		printf("%d%s",st->sudoku[i],(i%st->size==st->size-1) ? "\n" : " ");	
	}
	return PB_OK;
}

int subtree_check(subtree_t *st){
	int i,j,k; /*loop variables*/
	int l,m,n,size=st->size,bs=sqrt(size),row, col, block;
	int lut[size];
	

	for(i=0;i<size;i++)
		lut[i]=i/bs*size + i%bs;

	row = st->pos[st->clevel]/size;
	col = st->pos[st->clevel]%size;
	block = row/bs*bs + col/bs;
	for(k=1;k<=size;k++){
		l=m=n=2;
		for(j=0;j<size;j++){
			if (st->sudoku[row*size+j] ==k){
				l--;
				if (l==0) return(0);
			}
			if (st->sudoku[col+j*size] ==k){
				m--;
				if (m==0) return(0);
			}
			if (st->sudoku[lut[block]*bs + lut[j]] ==k){
				n--;
				if (n==0) return(0);
			}
		}
	}
	return(1);
}
/*works on the subtree, returns PB_EMPTY,PB_SOLVED*/
int subtree_work(subtree_t *st){
	int  valid,empty;

	valid = subtree_check(st);
	empty = isempty(st);	
	/*check	for termination conditions*/
	if(valid == 1 && st->clevel == st->free - 1)
		return PB_SOLVED;
	if(empty == 1)
		return PB_EMPTY;

	/*work*/
	if(valid){
		next_level(st);
	} else if(possibilities(st,st->clevel) > 0){
		advance(st);
	} else {
		return backtrack(st);
	}
	return PB_OK;
}

/*splits the problem in two*/
int subtree_split(subtree_t *stin, subtree_t *stout){
	int i,spltlvl,probwidth;

	if(stin->size != stout->size){
		printf("Error in subtree_split: incompatible sizes\n");
		return PB_ERROR;
	}

	if(stin->free != stout->free){
		printf("Error in subtree_split: incompatible number of free fields: in: %d vs out: %d\n",stin->free,stout->free);
		return PB_ERROR;
	}

	/*copy action*/

	for(i=0;i<stin->size*stin->size;i++)
		stout->sudoku[i] = stin->sudoku[i];

	for(i=0;i<stin->free;i++){
		stout->pos[i] = stin->pos[i];
		stout->upper[i] = stin->upper[i];
		stout->lower[i] = stin->lower[i];
	}

	/*split action*/
	spltlvl = 0;
	while(possibilities(stin,spltlvl) == 0){
		spltlvl++;
		if(spltlvl == stin->free){
			printf("Warning in subtree_split: atomic subproblem\n");
			return PB_EMPTY;
		}
	}

	if(possibilities(stin,spltlvl) > stin->size/2){
		probwidth = stin->size/2;
	} else {
		probwidth = possibilities(stin,spltlvl);
	}

	stout->upper[spltlvl] = stin->upper[spltlvl];
	stin->upper[spltlvl] -= probwidth;
	stout->lower[spltlvl] = stin->upper[spltlvl];

	stout->sudoku[stout->pos[spltlvl]] = stout->lower[spltlvl];

	for(i=spltlvl+1;i<stout->free;i++)
		stout->sudoku[stout->pos[i]] = 0;

	stout->clevel = spltlvl;

	return PB_OK;

}
